import sys

from copy import deepcopy
from inspect import isfunction

from pylons import config
from tg.controllers import TGController

rootcon = None

class entry(object):
    """
    This class contains all of the pieces of information about a given menu
    entry. It is the only public portion of this particular file. Everything
    else in here is private, and should not be used by any external code.
    """
    def __init__(self, path, name, extension, permission, url, extras={}, sortorder=999999, right=False, icon=None):
        self._path = path
        self._mpath = [x.strip() for x in path.split('||')]
        self._name = name
        self._extension = extension
        self._permission = permission
        self._url = url
        self.func = None
        self.base = None
        self.extras = deepcopy(extras)
        self.sortorder = sortorder
        self.icon = icon
        
        if 'class' not in self.extras:
            self.extras['class'] = []
            
        if not(hasattr(self.extras['class'], 'append')):
            self.extras['class'] = [self.extras['class'],]
        if right:            
            self.extras['class'].append('right')
        
    def __str__(self):
        return '%s at %s' % (str(self._path), str(self._url))


class shared_menu_cache(object):
    """
    This uses the Borg pattern to ensure a single menu cache across all
    requests. It completely encapsulates the shared menu cache, and there
    should be absolutely no reason to access the shared state directly.
    """
    __shared_state = {}
    def __init__(self):
        self.__dict__ = self.__shared_state
        if not(hasattr(self, '_menuitems')):
            self._menuitems = {}
        if not(hasattr(self, 'needs_update')):
            self.needs_update = True

    def addEntry(self, menuitem):
        self.needs_update = True
        if menuitem._name not in self._menuitems:
            self._menuitems[menuitem._name] = {}
        self._menuitems[menuitem._name][menuitem._path] = menuitem
        
    def removeEntry(self, menuitem):
        if menuitem._name in self._menuitems:
            if menuitem._path in self._menuitems[menuitem._name]:
                del self._menuitems[menuitem._name][menuitem._path]
            
    def getEntry(self, menuname, path):
        if menuname in self._menuitems:
            if path in self._menuitems[menuname]:
                return self._menuitems[menuname][path]
            else:
                return None
        else:
            return None
    
    def removeMenu(self, menuname):
        if menuname in self._menuitems:
            del self._menuitems[menuname]
            
    def getMenu(self, menuname):
        if self.needs_update:
            self.updateUrls()
        return self._menuitems[menuname] if menuname in self._menuitems else dict()
    
    def updateUrls(self):
        global rootcon
        if not rootcon:
            pname = '%s.controllers.root' % (config['package'].__name__)
            __import__(pname)
            rootcon = sys.modules[pname].RootController
        for menuname in self._menuitems:
            for menuitem in self._menuitems[menuname]:
                mi = self._menuitems[menuname][menuitem]
                mi._url = find_url(rootcon, self._menuitems[menuname][menuitem])
                if mi._url and mi._extension:
                    mi._url = '%s.%s' % (mi._url, mi._extension)
        self.needs_update = False


def find_url(root, menuitem):
    """
    This function is fairly complex, similar to permission_met. However, the
    goal is different: This function starts at the root of the menu tree,
    looking for a given function or class, so that it can return the path for
    that class starting at the root. The one exception is that it returns
    right away if it finds a ':' in the name, as it assumes an absolute URL.
    
    This method is recursive, so be aware of that as you trace through the
    code.
    """
    if ':' in str(menuitem._url):
        return menuitem._url
    if not(menuitem.base):
        if hasattr(menuitem.func, '__name__'):
            n = menuitem.func.__name__
            if n in root.__dict__:
                if root.__dict__[n] is menuitem.func:
                    return '/%s' % (n)
            for i in root.__dict__:
                controller = root.__dict__[i]
                if hasattr(controller, '_dispatch'):
                    v = find_url(controller.__class__, menuitem)
                    if v is not None:
                        return '/%s%s' % (i, v)
        else:
            return str(menuitem._url)
    else:
        for i in root.__dict__:
            controller = root.__dict__[i]
            if controller is menuitem.base:
                return '/%s/%s' % (i, menuitem._url)
            if hasattr(controller, '_dispatch'):
                v = find_url(controller.__class__, menuitem)
                if v is not None:
                    return '/%s%s' %(i, v)
    return None

##############################################################################
## Section: Functions to register and de-register user defined callbacks
##############################################################################

def register_callback(menuname, func):
    if menuname not in callbacks:
        callbacks[menuname] = []
    
    if func not in callbacks[menuname]:
        callbacks[menuname].append(func)

def register_callback_navbar(func):
    register_callback(u'navbar', func)

def register_callback_sidebar(func):
    register_callback(u'sidebar', func)
    
def deregister_callback(menuname, func):
    if menuname not in callbacks:
        return
    
    if func not in callbacks[menuname]:
        return
    
    callbacks[menuname].remove(func)

def deregister_callback_navbar(func):
    deregister_callback(u'navbar', func)

def deregister_callback_sidebar(func):
    deregister_callback(u'sidebar', func)
    
##############################################################################
## Section: Shared state variables, used to hold the menus and user-defined
## callbacks.
##############################################################################

shared_cache = shared_menu_cache()
callbacks = {'navbar': [], 'sidebar': []}


from copy import deepcopy
from sys import _getframe

from mako.template import Template

from pkg_resources import Requirement, resource_string
divmenu = Template(resource_string(Requirement.parse("tgext.menu"),"tgext/menu/templates/divmenu.mak"))

from pylons import config

from tgext.menu.caches import shared_cache, entry, callbacks
from tgext.menu.util import init_resources, OutputEntry
from tgext.menu.util import sort_entry, permission_met

jquery_bgiframe_js = jquery_dimensions_js = jquery_jdmenu_js = jquery_js = jquery_position_js = jquery_jdmenu_css = sortorder = None
use_tw2 = False

##############################################################################
## Section: url_from_menu, get_entry, switch_template
##############################################################################
def url_from_menu(menuname, menupath):
    menu = shared_cache.getMenu(menuname)
    menupath = [x.strip() for x in menupath.split('||')]
    for item in menu:
        if menu[item]._mpath == menupath:
            return menu[item]._url
    return None

def get_entry(menuname, menupath):
    return shared_cache.getEntry(menuname, menupath)

def switch_template(instr):
    global divmenu
    divmenu = Template(instr)

##############################################################################
## Section: Functions to append menu entries onto a menu
##############################################################################
def menu_append(path, name, extension=None, permission=None, url=None, base=None, extras={}, sortorder=999999, right=False, icon=None):
    item = entry(path, name, extension, permission, url, extras, sortorder, right, icon)
    item.base = base
    shared_cache.addEntry(item)

def navbar_append(path, extension=None, permission=None, url=None, base=None, extras={}, sortorder=999999, right=False, icon=None):
    menu_append(path, u'navbar', extension, permission, url, base, extras, sortorder, right, icon)

def sidebar_append(path, extension=None, permission=None, url=None, base=None, extras={}, sortorder=999999, right=False, icon=None):
    menu_append(path, u'sidebar', extension, permission, url, base, extras, sortorder, right, icon)

##############################################################################
## Section: Functions to remove menu entries from a menu
##############################################################################
def menu_remove(path, name):
    item = entry(path, name, None, None, None)
    shared_cache.removeEntry(item)

def navbar_remove(path):
    menu_remove(path, u'navbar')

def sidebar_remove(path):
    menu_remove(path, u'sidebar')

##############################################################################
## Section: Do final prep work to render a menu, and then call Mako to do the
## actual rendering
##############################################################################
def render_menu(menuname, vertical=False, active=None):
    global jquery_bgiframe_js, jquery_dimensions_js, jquery_jdmenu_js, jquery_js, jquery_position_js, jquery_jdmenu_css, sortorder, use_tw2
    
    if jquery_jdmenu_js is None:
        init_resources()


    if use_tw2:
        if config.get('tgext_menu', {}).get('inject_js', True):
            jquery_jdmenu_js.prepare()
    
        if config.get('tgext_menu', {}).get('inject_css', False):
            jquery_jdmenu_css.prepare()
    else:
        if config.get('tgext_menu', {}).get('inject_js', True):
            jquery_js.inject()
            jquery_bgiframe_js.inject()
            jquery_dimensions_js.inject()
            jquery_position_js.inject()
            jquery_jdmenu_js.inject()
        
        if config.get('tgext_menu', {}).get('inject_css', False):
            jquery_jdmenu_css.inject()
        
    menutree = OutputEntry(menuname)
    menu = shared_cache.getMenu(menuname)
    shortmenu = [menu[key] for key in filter(lambda x: permission_met(menu[x]), menu.keys())]
    if menuname in callbacks:
        for callback in callbacks[menuname]:
            try:
                shortmenu.extend(callback(menuname))
            except:
                pass
    if active:
        splitpath = [x.strip() for x in active.split('||')]
    else:
        splitpath = None
    for menuitem in sorted(shortmenu, sort_entry):
        if menuitem._mpath == splitpath:
            extras = deepcopy(menuitem.extras)
            extras['class'].append('active')
        else:
            extras = menuitem.extras
        menutree.appendPath(menuitem._mpath, str(menuitem._url), extras, menuitem.icon)
    return divmenu.render(menulist=menutree, name=menuname, vertical_menu=vertical)

def render_navbar(vertical=False, active=None):
    return render_menu(u'navbar', vertical, active)

def render_sidebar(vertical=False, active=None):
    return render_menu(u'sidebar', vertical, active)




import sys
from copy import deepcopy

from pylons import config
from tg import request

from repoze.what.predicates import NotAuthorizedError
from repoze.what.predicates import has_permission

import tgext.menu.functions

use_tw2 = False
jquery_bgiframe_js = None
jquery_dimensions_js = None
jquery_position_js = None
jquery_jdmenu_js = None
jquery_jdmenu_css = None
jquery_js = None
sortorder = {}

rootcon = None

class OutputEntry(object):
    """
    An internal class, used to represent a single entry's information that is
    necessary to render it.
    
    This class is used instead of the list method that was present in earlier
    revisions as this creates a tree structure from the list data. The result
    is something that is easy to display, easy to sort, and easy to work with
    at display time. Note that the list representation is left alone earlier
    as a way to make life easier during menu manipulation time. The two
    classes complement each other.
    """
    def __init__(self, name, href=None, extras={}, icon=None):
        self.children = []
        self.href = href
        self.name = name
        self.extras = deepcopy(extras)
        self.icon = icon
        
        if 'class' not in self.extras:
            self.extras['class'] = []

    def appendPath(self, name, href, extras={}, icon=None):
        # name is a list of paths, i.e.: ['Foo Spot', 'Bar']
        if len(name) > 1:
            try:
                idx = self.children.index(name[0])
            except ValueError:
                child = OutputEntry(name[0], extras=extras, icon=icon)
                self.children.append(child)
                idx = len(self.children)-1
            self.children[idx].appendPath(name[1:], href, extras, icon)
        else:
            child = OutputEntry(name[0], href, extras=extras, icon=icon)
            self.children.append(child)
    
    def __eq__(self, othername):
        return self.name == othername

def init_resources():
    """
    This method is reponsible for configuring the JSLink and CSSLink widgets
    as necessary. It automatically detects and uses ToscaWidgets 2 when
    available and possible. Outside of that, it uses ToscaWidgets 1.
    """
    global use_tw2, jquery_bgiframe_js, jquery_dimensions_js, jquery_position_js, jquery_jdmenu_js, jquery_jdmenu_css, jquery_js
    global sortorder

    sortorder = config.get('tgext_menu', {}).get('sortorder', {})

    use_tw2 = config.get('use_toscawidgets2', False)
    if use_tw2:
        from tw2.core import JSLink, CSSLink
        from tw2.jquery import jquery_js
                
        jquery_bgiframe_js = JSLink(
            modname=__name__,
            filename='static/js/jquery.bgiframe.js',
            resources=[jquery_js])
        
        jquery_dimensions_js = JSLink(
            modname=__name__,
            filename='static/js/jquery.dimensions.js',
            resources=[jquery_js])
        
        jquery_position_js = JSLink(
            modname=__name__,
            filename='static/js/jquery.positionBy.js',
            resources=[jquery_js, jquery_bgiframe_js, jquery_dimensions_js])
        
        jquery_jdmenu_js = JSLink(
            modname=__name__,
            filename='static/js/jquery.jdMenu.js',
            resources=[jquery_js, jquery_bgiframe_js, jquery_dimensions_js, jquery_position_js]).req()
        
        jquery_jdmenu_css = CSSLink(modname=__name__, filename='static/css/jquery.jdMenu.css').req()
    else:
        from tw.api import JSLink, CSSLink
        from tw.jquery import jquery_js, jquery_bgiframe_js, jquery_dimensions_js
        
        jquery_position_js = JSLink(
            modname=__name__,
            filename='static/js/jquery.positionBy.js',
            javascript=[jquery_js, jquery_bgiframe_js, jquery_dimensions_js])
    
        jquery_jdmenu_js = JSLink(
            modname=__name__,
            filename='static/js/jquery.jdMenu.js',
            javascript=[jquery_js, jquery_bgiframe_js, jquery_dimensions_js, jquery_position_js])
    
        jquery_jdmenu_css = CSSLink(modname=__name__, filename='static/css/jquery.jdMenu.css')
        
    tgext.menu.functions.sortorder = sortorder
    tgext.menu.functions.jquery_bgiframe_js = jquery_bgiframe_js
    tgext.menu.functions.jquery_dimensions_js = jquery_dimensions_js
    tgext.menu.functions.jquery_jdmenu_css = jquery_jdmenu_css
    tgext.menu.functions.jquery_jdmenu_js = jquery_jdmenu_js
    tgext.menu.functions.jquery_js = jquery_js
    tgext.menu.functions.jquery_position_js = jquery_position_js
    tgext.menu.functions.use_tw2 = use_tw2

def sort_entry(ent1, ent2):
    """
    A simple comparison function. It uses the sort order specified in the
    application's configuration, compares the menu entries, and returns the
    correct value depending on the configuration.
    """
    global sortorder
    idx = 0
    while idx < len(ent1._mpath) and idx < len(ent2._mpath):
        val1 = sortorder.get(ent1._mpath[idx], ent1.sortorder)
        val2 = sortorder.get(ent2._mpath[idx], ent2.sortorder)
        if val1 != val2:
            return cmp(val1, val2)
        idx += 1
    return cmp(ent1._mpath, ent2._mpath)

def permission_met(menu):
    """
    This is one of the more complicated methods. It works recursively.
    
    When called, it is given the root of the controller hierarchy. It looks
    for the path to the menu entry, and checks everything that it can along
    the way: allow_only on all controllers, and the (optional) permission on
    the method itself (which must be given to the @menu decorator or
    menu_append, see the README for details why and a workaround).
    """
    global rootcon
    retval = True
    
    if not rootcon:
        pname = '%s.controllers.root' % (config['package'].__name__)
        __import__(pname)
        rootcon = sys.modules[pname].RootController

    # Check to see if specific menu permission has been set
    permission = menu._permission
    if type(permission) is str:
        try:
            has_permission(permission).check_authorization(request.environ)
            return True
        except NotAuthorizedError:
            return False
    elif permission is not None:
        try:
            permission.check_authorization(request.environ)
            return True
        except:
            return False
    else:
        # No specific menu permission has been set, walk the tree
        lpath = menu._url.split('/')[1:]
        currcon = rootcon
        for component in lpath:
            if hasattr(currcon, 'allow_only'):
                try:
                    getattr(currcon, 'allow_only').check_authorization(request.environ)
                except:
                    return False
            if hasattr(currcon, component):
                currcon = getattr(currcon, component)
            else:
                break
        
    return True




"""
tgext.menu is a tool to allow the web developer to specify where a given
method should appear in the menu hierarchy without having to deal with
templates. For how to use this, please see the README file. As this is geared
towards developers, the README file is the complete reference.

Copyright 2010 Michael J. Pedersen <m.pedersen@icelus.org>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including witho ut limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the fol lowing conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import sys

from tg.configuration import Bunch
from pylons import config

from caches import register_callback, register_callback_navbar, register_callback_sidebar
from caches import deregister_callback, deregister_callback_navbar, deregister_callback_sidebar
from caches import entry
from decorators import menu, navbar, sidebar
from functions import render_menu, render_navbar, render_sidebar
from functions import menu_append, navbar_append, sidebar_append
from functions import menu_remove, navbar_remove, sidebar_remove
from functions import url_from_menu, get_entry, switch_template

__all__ = ["menu", "navbar", "sidebar",
           "menu_append", "navbar_append", "sidebar_append",
           "menu_remove", "navbar_remove", "sidebar_remove",
           "render_menu", "render_navbar", "render_sidebar",
           "register_callback", "register_callback_navbar", "register_callback_sidebar",
           "deregister_callback", "deregister_callback_navbar", "deregister_callback_sidebar",
           "url_from_menu", "get_entry", "switch_template", "entry"]

# The below code is a bit of magic to make the usage of tgext.menu easier on
# the developer. Basically, it grabs the current variable_provider from the
# config, puts the tgext.menu variable provider into the config, and then
# returns the results from its own plus the original variable provider.
#
# By doing this, the developer does not have to remember to update the
# variable provider to provide the functions that the README says should be
# available in all templates.
def menu_variable_provider():
    menu_vars = Bunch (
        url_from_menu = url_from_menu,
        render_menu = render_menu,
        render_navbar = render_navbar,
        render_sidebar = render_sidebar,
    )
    
    try:
        from genshi import HTML
        menu_vars['HTML'] = HTML
    except ImportError:
        pass
    
    if app_variable_provider:
        menu_vars.update(app_variable_provider())

    return menu_vars

app_variable_provider = config.get('tgext_menu_sub_variable_provider', None)
config.update(Bunch(
    variable_provider = menu_variable_provider
    ))



# -*- coding: utf-8 -*-
try:
    from setuptools import setup, find_packages
except ImportError:
    from ez_setup import use_setuptools
    use_setuptools()
    from setuptools import setup, find_packages

import os

setup(
    name='tgext.menu',
    version='1.0b3',
    description='Automatic menu/navbar/sidebar generation extension for TurboGears',
    author='Michael Pedersen',
    author_email='m.pedersen@icelus.org',
    url='http://bitbucket.org/pedersen/tgext.menu',
    install_requires=[
        "TurboGears2 >= 2.1b1",
        'Mako >= 0.2.4',
        ],
    setup_requires=["PasteScript >= 1.7"],
    paster_plugins=['PasteScript', 'Pylons', 'TurboGears2', 'tg.devtools'],
    packages=find_packages(exclude=['ez_setup', "*.test", "*.test.*", "test.*", "test"]),
    namespace_packages=['tgext'],
    include_package_data=True,
    test_suite='nose.collector',
    tests_require=['WebTest', 'BeautifulSoup', "zope.sqlalchemy >= 0.4 ", "tw2.jquery", "tw2.forms", "tw2.core"],
    message_extractors={'tgextmenu': [
            ('**.py', 'python', None),
            ('templates/**.mako', 'mako', None),
            ('templates/**.html', 'genshi', None),
            ('public/**', 'ignore', None)]},
    dependency_links=['http://www.turbogears.org/2.1/downloads/current/index'],
    zip_safe=False
)
